"""Holds the reference materials for Tidy3D material library."""

import pydantic.v1 as pd

from ..components.base import Tidy3dBaseModel


class ReferenceData(Tidy3dBaseModel):
    """Reference data."""

    doi: str = pd.Field(None, title="DOI", description="DOI of the reference.")
    journal: str = pd.Field(
        None,
        title="Journal publication info",
        description="Publication info in the order of author, title, journal volume, and year.",
    )
    url: str = pd.Field(
        None,
        title="URL link",
        description="Some reference can be accessed through a url link to its pdf etc.",
    )


material_refs = dict(
    Li2014=ReferenceData(
        journal="Y. Li, A. Chernikov, X. Zhang, A. Rigosi, H. M. Hill, A. M. van der Zande, "
        "D. A. Chenet, E. Shih, J. Hone, and T. F. Heinz. Measurement of the optical dielectric "
        "function of monolayer transition-metal dichalcogenides: MoS2, MoSe2, WS2, and WSe2, "
        "Phys. Rev. B 90, 205422 (2014)",
        doi="https://doi.org/10.1103/PhysRevB.90.205422",
    ),
    Yang2015=ReferenceData(
        journal="H. U. Yang, J. D'Archangel, M. L. Sundheimer, E. Tucker, G. D. Boreman, "
        "M. B. Raschke. Optical dielectric function of silver, Phys. Rev. B 91, 235137 (2015)",
        doi="https://journals.aps.org/prb/abstract/10.1103/PhysRevB.91.235137",
    ),
    Olmon2012=ReferenceData(
        journal="R. L. Olmon, B. Slovick, T. W. Johnson, D. Shelton, S.-H. Oh, "
        "G. D. Boreman, and M. B. Raschke. Optical dielectric function of "
        "gold, Phys. Rev. B 86, 235147 (2012)",
        doi="https://doi.org/10.1103/PhysRevB.86.235147",
    ),
    Rakic1995=ReferenceData(
        journal="A. D. Rakic. Algorithm for the determination of intrinsic optical "
        "constants of metal films: application to aluminum, Appl. Opt. 34, 4755-4767 (1995)",
        doi="https://doi.org/10.1364/AO.34.004755",
    ),
    Rakic1998=ReferenceData(
        journal="A. D. Rakic, A. B. Djurisic, J. M. Elazar, and M. L. Majewski. "
        "Optical properties of metallic films for vertical-cavity optoelectronic "
        "devices, Appl. Opt. 37, 5271-5283 (1998)",
        doi="https://doi.org/10.1364/AO.37.005271",
    ),
    JohnsonChristy1972=ReferenceData(
        journal="P. B. Johnson and R. W. Christy. Optical constants of the noble "
        "metals, Phys. Rev. B 6, 4370-4379 (1972)",
        doi="https://doi.org/10.1103/PhysRevB.6.4370",
    ),
    Horiba=ReferenceData(
        journal="Horiba Technical Note 08: Lorentz Dispersion Model",
        url="http://www.horiba.com/fileadmin/uploads/Scientific/Downloads"
        "/OpticalSchool_CN/TN/ellipsometer/Lorentz_Dispersion_Model.pdf",
    ),
    FernOnton1971=ReferenceData(
        journal="R. E. Fern and A. Onton. Refractive index of AlAs, "
        "J. Appl. Phys. 42, 3499-3500 (1971)",
        doi="https://doi.org/10.1063/1.1660760",
    ),
    Sultanova2009=ReferenceData(
        journal="N. Sultanova, S. Kasarova and I. Nikolov. Dispersion properties "
        "of optical polymers, Acta Physica Polonica A 116, 585-587 (2009)",
        doi="https://doi.org/10.12693/aphyspola.116.585",
    ),
    Malitson1965=ReferenceData(
        journal="I. H. Malitson. Interspecimen comparison of the refractive "
        "index of fused silica, J. Opt. Soc. Am. 55, 1205-1208 (1965)",
        doi="https://doi.org/10.1364/JOSA.55.001205",
    ),
    Tan1998=ReferenceData(
        journal="C. Z. Tan. Determination of refractive index of silica glass "
        "for infrared wavelengths by IR spectroscopy, J. Non-Cryst. Solids 223, 158-163 (1998)",
        doi="https://doi.org/10.1016/S0022-3093(97)00438-9",
    ),
    Skauli2003=ReferenceData(
        journal="T. Skauli, P. S. Kuo, K. L. Vodopyanov, T. J. Pinguet, "
        "O. Levi, L. A. Eyres, J. S. Harris, M. M. Fejer, B. Gerard, "
        "L. Becouarn, and E. Lallier. Improved dispersion relations "
        "for GaAs and applications to nonlinear optics, J. Appl. Phys., 94, 6447-6455 (2003)",
        doi="https://doi.org/10.1063/1.1621740",
    ),
    Icenogle1976=ReferenceData(
        journal="H. W. Icenogle, Ben C. Platt, and William L. Wolfe. "
        "Refractive indexes and temperature coefficients of germanium "
        "and silicon Appl. Opt. 15 2348-2351 (1976)",
        doi="https://doi.org/10.1364/AO.15.002348",
    ),
    Barnes1979=ReferenceData(
        journal="N. P. Barnes and M. S. Piltch. Temperature-dependent "
        "Sellmeier coefficients and nonlinear optics average power limit "
        "for germanium J. Opt. Soc. Am. 69 178-180 (1979)",
        doi="https://doi.org/10.1364/JOSA.69.000178",
    ),
    Pettit1965=ReferenceData(
        journal="G. D. Pettit and W. J. Turner. Refractive index of InP, "
        "J. Appl. Phys. 36, 2081 (1965)",
        doi="https://doi.org/10.1063/1.1714410",
    ),
    Pikhtin1978=ReferenceData(
        journal="A. N. Pikhtin and A. D. Yas'kov. Disperson of the "
        "refractive index of semiconductors with diamond and zinc-blende "
        "structures, Sov. Phys. Semicond. 12, 622-626 (1978)",
    ),
    HandbookOptics=ReferenceData(
        journal="Handbook of Optics, 2nd edition, Vol. 2. McGraw-Hill 1994 (ISBN 9780070479746)",
    ),
    StephensMalitson1952=ReferenceData(
        journal="R. E. Stephens and I. H. Malitson. Index of refraction of "
        "magnesium oxide, J. Res. Natl. Bur. Stand. 49 249-252 (1952)",
        doi="https://doi.org/10.6028/jres.049.025",
    ),
    Werner2009=ReferenceData(
        journal="W. S. M. Werner, K. Glantschnig, C. Ambrosch-Draxl. "
        "Optical constants and inelastic electron-scattering data for 17 "
        "elemental metals, J. Phys Chem Ref. Data 38, 1013-1092 (2009)",
        doi="https://doi.org/10.1063/1.3243762",
    ),
    Luke2015=ReferenceData(
        journal="K. Luke, Y. Okawachi, M. R. E. Lamont, A. L. Gaeta, M. Lipson. "
        "Broadband mid-infrared frequency comb generation in a Si3N4 microresonator, "
        "Opt. Lett. 40, 4823-4826 (2015)",
        doi="https://doi.org/10.1364/OL.40.004823",
    ),
    Philipp1973=ReferenceData(
        journal="H. R. Philipp. Optical properties of silicon nitride, "
        "J. Electrochim. Soc. 120, 295-300 (1973)",
        doi="https://doi.org/10.1149/1.2403440",
    ),
    Baak1982=ReferenceData(
        journal="T. Baak. Silicon oxynitride; a material for GRIN optics, "
        "Appl. Optics 21, 1069-1072 (1982)",
        doi="https://doi.org/10.1364/AO.21.001069",
    ),
    Nigara1968=ReferenceData(
        journal="Y. Nigara. Measurement of the optical constants of yttrium oxide, "
        "Jpn. J. Appl. Phys. 7, 404-408 (1968)",
        doi="https://doi.org/10.1143/JJAP.7.404",
    ),
    Zelmon1997=ReferenceData(
        journal="D. E. Zelmon, D. L. Small and D. Jundt. Infrared corrected Sellmeier "
        "coefficients for congruently grown lithium niobate and 5 mol.% magnesium oxide-doped "
        "lithium niobate, J. Opt. Soc. Am. B 14, 3319-3322 (1997)",
        doi="https://doi.org/10.1364/JOSAB.14.003319",
    ),
    Zelmon1998=ReferenceData(
        journal="D. E. Zelmon, D. L. Small and R. Page. Refractive-index measurements "
        "of undoped yttrium aluminum garnet from 0.4 to 5.0 μm, Appl. Opt. 37, 4933-4935 (1998)",
        doi="https://doi.org/10.1364/AO.37.004933",
    ),
    SalzbergVilla1957=ReferenceData(
        journal="C. D. Salzberg and J. J. Villa. Infrared Refractive Indexes of "
        "Silicon, Germanium and Modified Selenium Glass, J. Opt. Soc. Am., 47, 244-246 (1957)",
        doi="https://doi.org/10.1364/JOSA.47.000244",
    ),
    Tatian1984=ReferenceData(
        journal="B. Tatian. Fitting refractive-index data with the Sellmeier "
        "dispersion formula, Appl. Opt. 23, 4477-4485 (1984)",
        doi="https://doi.org/10.1364/AO.23.004477",
    ),
    Li1993_293K=ReferenceData(
        journal="H. H. Li. Refractive index of silicon and germanium and its wavelength "
        "and temperature derivatives, J. Phys. Chem. Ref. Data 9, 561-658 (1993)",
        doi="https://doi.org/10.1063/1.555624",
    ),
    Green2008=ReferenceData(
        journal="M. A. Green. Self-consistent optical parameters of intrinsic silicon "
        "at 300K including temperature coefficients, Sol. Energ. Mat. "
        "Sol. Cells 92, 1305–1310 (2008)",
        doi="https://doi.org/10.1016/j.solmat.2008.06.009",
    ),
    Zemax=ReferenceData(
        journal="SCHOTT Zemax catalog 2017-01-20b",
        url="https://refractiveindex.info/download/data/2017/schott_2017-01-20.pdf",
    ),
    Hanson2008=ReferenceData(
        journal="George W. Hanson. Dyadic Green's Functions for an Anisotropic, "
        "Non-Local Model of Biased Graphene, IEEE Trans. Antennas Propag. 56, 3, 747-757 (2008)",
        doi="https://doi.org/10.1109/TAP.2008.917005",
    ),
    Burnett2016=ReferenceData(
        journal="John H. Burnett, Simon G. Kaplan, Eric Stover, and Adam Phenis, "
        "Refractive index measurements of Ge, "
        "Proc. SPIE 9974, Infrared Sensors, Devices, and Applications VI, 99740X "
        "(20 September 2016)",
        doi="https://doi.org/10.1117/12.2237978",
    ),
    Palik=ReferenceData(
        journal="E. D. Palik. Handbook of Optical Constants of Solids, " "Academic Press (1998)",
        doi="https://doi.org/10.1016/B978-0-08-055630-7.50001-8",
    ),
    Palik_Lossy=ReferenceData(
        journal="E. D. Palik. Handbook of Optical Constants of Solids, " "Academic Press (1998)",
        doi="https://doi.org/10.1016/B978-0-08-055630-7.50001-8",
    ),
    Palik_Lossless=ReferenceData(
        journal="E. D. Palik. Handbook of Optical Constants of Solids, " "Academic Press (1998)",
        doi="https://doi.org/10.1016/B978-0-08-055630-7.50001-8",
    ),
)



from typing import Iterable, List, Literal, Tuple, Union

import autograd.numpy as np
from autograd.scipy.signal import convolve as convolve_ag
from numpy.typing import NDArray

from tidy3d.components.autograd.functions import interpn

from .types import PaddingType

__all__ = [
    "interpn",
    "pad",
    "convolve",
    "grey_dilation",
    "grey_erosion",
    "grey_opening",
    "grey_closing",
    "morphological_gradient",
    "morphological_gradient_internal",
    "morphological_gradient_external",
    "rescale",
    "threshold",
]


def _make_slices(rule: Union[int, slice], ndim: int, axis: int) -> Tuple[slice, ...]:
    """Create a tuple of slices for indexing an array.

    Parameters
    ----------
    rule : Union[int, slice]
        The rule to apply on the specified axis.
    ndim : int
        The number of dimensions of the array.
    axis : int
        The axis to which the rule should be applied.

    Returns
    -------
    Tuple[slice, ...]
        A tuple of slices for indexing.
    """
    return tuple(slice(None) if d != axis else rule for d in range(ndim))


def _constant_pad(
    array: NDArray, pad_width: Tuple[int, int], axis: int, *, constant_value: float = 0.0
) -> NDArray:
    """Pad an array with a constant value along a specified axis.

    Parameters
    ----------
    array : NDArray
        The input array to pad.
    pad_width : Tuple[int, int]
        The number of values padded to the edges of each axis.
    axis : int
        The axis along which to pad.
    constant_value : float = 0.0
        The constant value to pad with.

    Returns
    -------
    NDArray
        The padded array.
    """
    p = [(pad_width[0], pad_width[1]) if ax == axis else (0, 0) for ax in range(array.ndim)]
    return np.pad(array, p, mode="constant", constant_values=constant_value)


def _edge_pad(array: NDArray, pad_width: Tuple[int, int], axis: int) -> NDArray:
    """Pad an array using the `edge` mode along a specified axis.

    Parameters
    ----------
    array : NDArray
        The input array to pad.
    pad_width : Tuple[int, int]
        The number of values padded to the edges of each axis.
    axis : int
        The axis along which to pad.

    Returns
    -------
    NDArray
        The padded array.
    """
    left, right = (_make_slices(rule, array.ndim, axis) for rule in (0, -1))

    cat_arys = []
    if pad_width[0] > 0:
        cat_arys.append(np.stack(pad_width[0] * [array[left]], axis=axis))
    cat_arys.append(array)
    if pad_width[1] > 0:
        cat_arys.append(np.stack(pad_width[1] * [array[right]], axis=axis))

    return np.concatenate(cat_arys, axis=axis)


def _reflect_pad(array: NDArray, pad_width: Tuple[int, int], axis: int) -> NDArray:
    """Pad an array using the `reflect` mode along a specified axis.

    Parameters
    ----------
    array : NDArray
        The input array to pad.
    pad_width : Tuple[int, int]
        The number of values padded to the edges of each axis.
    axis : int
        The axis along which to pad.

    Returns
    -------
    NDArray
        The padded array.
    """
    left, right = (
        _make_slices(rule, array.ndim, axis)
        for rule in (slice(pad_width[0], 0, -1), slice(-2, -pad_width[1] - 2, -1))
    )
    return np.concatenate([array[left], array, array[right]], axis=axis)


def _symmetric_pad(array: NDArray, pad_width: Tuple[int, int], axis: int) -> NDArray:
    """Pad an array using the `symmetric` mode along a specified axis.

    Parameters
    ----------
    array : NDArray
        The input array to pad.
    pad_width : Tuple[int, int]
        The number of values padded to the edges of each axis.
    axis : int
        The axis along which to pad.

    Returns
    -------
    NDArray
        The padded array.
    """
    left, right = (
        _make_slices(rule, array.ndim, axis)
        for rule in (
            slice(pad_width[0] - 1, None, -1) if pad_width[0] > 0 else slice(0, 0),
            slice(-1, -pad_width[1] - 1, -1),
        )
    )
    return np.concatenate([array[left], array, array[right]], axis=axis)


def _wrap_pad(array: NDArray, pad_width: Tuple[int, int], axis: int) -> NDArray:
    """Pad an array using the `wrap` mode along a specified axis.

    Parameters
    ----------
    array : NDArray
        The input array to pad.
    pad_width : Tuple[int, int]
        The number of values padded to the edges of each axis.
    axis : int
        The axis along which to pad.

    Returns
    -------
    NDArray
        The padded array.
    """
    left, right = (
        _make_slices(rule, array.ndim, axis)
        for rule in (
            slice(-pad_width[0], None) if pad_width[0] > 0 else slice(0, 0),
            slice(0, pad_width[1]) if pad_width[1] > 0 else slice(0, 0),
        )
    )
    return np.concatenate([array[left], array, array[right]], axis=axis)


def pad(
    array: NDArray,
    pad_width: Union[int, Tuple[int, int]],
    *,
    mode: PaddingType = "constant",
    axis: Union[int, Iterable[int], None] = None,
    constant_value: float = 0.0,
) -> NDArray:
    """Pad an array along a specified axis with a given mode and padding width.

    Parameters
    ----------
    array : NDArray
        The input array to pad.
    pad_width : Union[int, Tuple[int, int]]
        The number of values padded to the edges of each axis. If an integer is provided,
        it is used for both the left and right sides. If a tuple is provided, it specifies
        the padding for the left and right sides respectively.
    mode : PaddingType = "constant"
        The padding mode to use.
    axis : Union[int, Iterable[int], None] = None
        The axis or axes along which to pad. If None, padding is applied to all axes.
    constant_value : float = 0.0
        The value to set the padded values for "constant" mode.

    Returns
    -------
    NDArray
        The padded array.

    Raises
    ------
    ValueError
        If the padding width has more than two elements or if padding is negative.
    NotImplementedError
        If padding larger than the input size is requested.
    KeyError
        If an unsupported padding mode is specified.
    IndexError
        If an axis is out of range for the array dimensions.

    Examples
    --------
    >>> import numpy as np
    >>> from tidy3d.plugins.autograd.functions import pad
    >>> array = np.array([[1, 2], [3, 4]])
    >>> pad(array, (1, 1), mode="constant", constant_value=0)
    array([[0, 0, 0, 0],
           [0, 1, 2, 0],
           [0, 3, 4, 0],
           [0, 0, 0, 0]])
    >>> pad(array, 1, mode="reflect")
    array([[4, 3, 4, 3],
           [2, 1, 2, 1],
           [4, 3, 4, 3],
           [2, 1, 2, 1]])
    """
    pad_width = np.atleast_1d(pad_width)

    if pad_width.size == 1:
        pad_width = np.array([pad_width[0], pad_width[0]])
    elif pad_width.size != 2:
        raise ValueError("Padding width must have one or two elements, got {pad_width.size}.")

    if any(any(p >= s for p in pad_width) for s in array.shape):
        raise NotImplementedError("Padding larger than the input size is not supported.")
    if any(p < 0 for p in pad_width):
        raise ValueError("Padding must be positive.")
    if all(p == 0 for p in pad_width):
        return array

    if axis is None:
        axis = range(array.ndim)

    _mode_map = {
        "constant": _constant_pad,
        "edge": _edge_pad,
        "reflect": _reflect_pad,
        "symmetric": _symmetric_pad,
        "wrap": _wrap_pad,
    }

    try:
        pad_fun = _mode_map[mode]
    except KeyError as e:
        raise KeyError(f"Unsupported padding mode: {mode}.") from e

    for ax in np.atleast_1d(axis):
        if ax < 0:
            ax += array.ndim
        if ax < 0 or ax >= array.ndim:
            raise IndexError(f"Axis out of range for array with {array.ndim} dimensions.")

        array = pad_fun(array, pad_width, axis=ax)

    return array


def convolve(
    array: NDArray,
    kernel: NDArray,
    *,
    padding: PaddingType = "constant",
    axes: Union[Tuple[List[int], List[int]], None] = None,
    mode: Literal["full", "valid", "same"] = "same",
) -> NDArray:
    """Convolve an array with a given kernel.

    Parameters
    ----------
    array : NDArray
        The input array to be convolved.
    kernel : NDArray
        The kernel to convolve with the input array. All dimensions of the kernel must be odd.
    padding : PaddingType = "constant"
        The padding mode to use.
    axes : Union[Tuple[List[int], List[int]], None] = None
        The axes along which to perform the convolution.
    mode : Literal["full", "valid", "same"] = "same"
        The convolution mode.

    Returns
    -------
    NDArray
        The result of the convolution.

    Raises
    ------
    ValueError
        If any dimension of the kernel is even.
        If the dimensions of the kernel do not match the dimensions of the array.
    """
    if any(k % 2 == 0 for k in kernel.shape):
        raise ValueError(f"All kernel dimensions must be odd, got {kernel.shape}.")

    if kernel.ndim != array.ndim and axes is None:
        raise ValueError(
            f"Kernel dimensions must match array dimensions, got kernel {kernel.shape} and array {array.shape}."
        )

    if mode in ("same", "full"):
        kernel_dims = kernel.shape if axes is None else [kernel.shape[d] for d in axes[1]]
        pad_widths = [(ks // 2, ks // 2) for ks in kernel_dims]
        for axis, pad_width in enumerate(pad_widths):
            array = pad(array, pad_width, mode=padding, axis=axis)
        mode = "valid" if mode == "same" else mode

    return convolve_ag(array, kernel, axes=axes, mode=mode)


def grey_dilation(
    array: NDArray,
    size: Union[Union[int, Tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Perform grey dilation on an array.

    Parameters
    ----------
    array : NDArray
        The input array to perform grey dilation on.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[NDArray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    NDArray
        The result of the grey dilation operation.

    Raises
    ------
    ValueError
        If both `size` and `structure` are None.
    """
    if size is None and structure is None:
        raise ValueError("Either size or structure must be provided.")

    if size is not None:
        size = np.atleast_1d(size)
        shape = (size[0], size[-1])
        nb = np.zeros(shape)
    elif np.all(structure == 0):
        nb = np.zeros_like(structure)
    else:
        nb = np.copy(structure)
        nb[structure == 0] = -maxval

    h, w = nb.shape
    bias = np.reshape(nb, (-1, 1, 1))
    kernel = np.reshape(np.eye(h * w), (h * w, h, w))

    array = convolve(array, kernel, axes=((0, 1), (1, 2)), padding=mode) + bias
    return np.max(array, axis=0)


def grey_erosion(
    array: NDArray,
    size: Union[Union[int, Tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Perform grey erosion on an array.

    Parameters
    ----------
    array : NDArray
        The input array to perform grey dilation on.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[NDArray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    NDArray
        The result of the grey dilation operation.

    Raises
    ------
    ValueError
        If both `size` and `structure` are None.
    """
    if size is None and structure is None:
        raise ValueError("Either size or structure must be provided.")

    if size is not None:
        size = np.atleast_1d(size)
        shape = (size[0], size[-1])
        nb = np.zeros(shape)
    elif np.all(structure == 0):
        nb = np.zeros_like(structure)
    else:
        nb = np.copy(structure)
        nb[structure == 0] = -maxval

    h, w = nb.shape
    bias = np.reshape(nb, (-1, 1, 1))
    kernel = np.reshape(np.eye(h * w), (h * w, h, w))

    array = convolve(array, kernel, axes=((0, 1), (1, 2)), padding=mode) - bias
    return np.min(array, axis=0)


def grey_opening(
    array: NDArray,
    size: Union[Union[int, Tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Perform grey opening on an array.

    Parameters
    ----------
    array : NDArray
        The input array to perform grey opening on.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[NDArray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    NDArray
        The result of the grey opening operation.
    """
    array = grey_erosion(array, size, structure, mode=mode, maxval=maxval)
    array = grey_dilation(array, size, structure, mode=mode, maxval=maxval)
    return array


def grey_closing(
    array: NDArray,
    size: Union[Union[int, Tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Perform grey closing on an array.

    Parameters
    ----------
    array : NDArray
        The input array to perform grey closing on.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[NDArray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    NDArray
        The result of the grey closing operation.
    """
    array = grey_dilation(array, size, structure, mode=mode, maxval=maxval)
    array = grey_erosion(array, size, structure, mode=mode, maxval=maxval)
    return array


def morphological_gradient(
    array: NDArray,
    size: Union[Union[int, Tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Compute the morphological gradient of an array.

    Parameters
    ----------
    array : NDArray
        The input array to compute the morphological gradient of.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[NDArray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    NDArray
        The morphological gradient of the input array.
    """
    return grey_dilation(array, size, structure, mode=mode, maxval=maxval) - grey_erosion(
        array, size, structure, mode=mode, maxval=maxval
    )


def morphological_gradient_internal(
    array: NDArray,
    size: Union[Union[int, Tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Compute the internal morphological gradient of an array.

    Parameters
    ----------
    array : NDArray
        The input array to compute the internal morphological gradient of.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[NDArray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    NDArray
        The internal morphological gradient of the input array.
    """
    return array - grey_erosion(array, size, structure, mode=mode, maxval=maxval)


def morphological_gradient_external(
    array: NDArray,
    size: Union[Union[int, Tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Compute the external morphological gradient of an array.

    Parameters
    ----------
    array : NDArray
        The input array to compute the external morphological gradient of.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[NDArray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    NDArray
        The external morphological gradient of the input array.
    """
    return grey_dilation(array, size, structure, mode=mode, maxval=maxval) - array


def rescale(
    array: NDArray, out_min: float, out_max: float, in_min: float = 0.0, in_max: float = 1.0
) -> NDArray:
    """
    Rescale an array from an arbitrary input range to an arbitrary output range.

    Parameters
    ----------
    array : NDArray
        The input array to be rescaled.
    out_min : float
        The minimum value of the output range.
    out_max : float
        The maximum value of the output range.
    in_min : float = 0.0
        The minimum value of the input range.
    in_max : float = 1.0
        The maximum value of the input range.

    Returns
    -------
    NDArray
        The rescaled array.
    """

    if in_min == in_max:
        raise ValueError(
            f"'in_min' ({in_min}) must not be equal to 'in_max' ({in_max}) "
            "to avoid division by zero."
        )
    if out_min >= out_max:
        raise ValueError(f"'out_min' ({out_min}) must be less than 'out_max' ({out_max}).")
    if in_min >= in_max:
        raise ValueError(f"'in_min' ({in_min}) must be less than 'in_max' ({in_max}).")

    scaled = (array - in_min) / (in_max - in_min)
    return scaled * (out_max - out_min) + out_min


def threshold(
    array: NDArray, vmin: float = 0.0, vmax: float = 1.0, level: Union[float, None] = None
) -> NDArray:
    """Apply a threshold to an array, setting values below the threshold to `vmin` and values above to `vmax`.

    Parameters
    ----------
    array : NDArray
        The input array to be thresholded.
    vmin : float = 0.0
        The value to assign to elements below the threshold.
    vmax : float = 1.0
        The value to assign to elements above the threshold.
    level : Union[float, None] = None
        The threshold level. If None, the threshold is set to the midpoint between `vmin` and `vmax`.

    Returns
    -------
    NDArray
        The thresholded array.
    """
    if vmin >= vmax:
        raise ValueError(
            f"Invalid threshold range: 'vmin' ({vmin}) must be smaller than 'vmax' ({vmax})."
        )

    if level is None:
        level = (vmin + vmax) / 2
    elif not (vmin <= level <= vmax):
        raise ValueError(
            f"Invalid threshold level: 'level' ({level}) must be "
            f"between 'vmin' ({vmin}) and 'vmax' ({vmax})."
        )

    return np.where(array < level, vmin, vmax)



"""Finite-difference derivatives and PML absorption operators expressed as sparse matrices."""

import numpy as np
import scipy.sparse as sp

from ...constants import EPSILON_0, ETA_0


def make_dxf(dls, shape, pmc):
    """Forward derivative in x."""
    Nx, Ny = shape
    if Nx == 1:
        return sp.csr_matrix((Ny, Ny))
    dxf = sp.csr_matrix(sp.diags([-1, 1], [0, 1], shape=(Nx, Nx)))
    if not pmc:
        dxf[0, 0] = 0.0
    dxf = sp.diags(1 / dls).dot(dxf)
    dxf = sp.kron(dxf, sp.eye(Ny))
    return dxf


def make_dxb(dls, shape, pmc):
    """Backward derivative in x."""
    Nx, Ny = shape
    if Nx == 1:
        return sp.csr_matrix((Ny, Ny))
    dxb = sp.csr_matrix(sp.diags([1, -1], [0, -1], shape=(Nx, Nx)))
    if pmc:
        dxb[0, 0] = 2.0
    else:
        dxb[0, 0] = 0.0
    dxb = sp.diags(1 / dls).dot(dxb)
    dxb = sp.kron(dxb, sp.eye(Ny))
    return dxb


def make_dyf(dls, shape, pmc):
    """Forward derivative in y."""
    Nx, Ny = shape
    if Ny == 1:
        return sp.csr_matrix((Nx, Nx))
    dyf = sp.csr_matrix(sp.diags([-1, 1], [0, 1], shape=(Ny, Ny)))
    if not pmc:
        dyf[0, 0] = 0.0
    dyf = sp.diags(1 / dls).dot(dyf)
    dyf = sp.kron(sp.eye(Nx), dyf)
    return dyf


def make_dyb(dls, shape, pmc):
    """Backward derivative in y."""
    Nx, Ny = shape
    if Ny == 1:
        return sp.csr_matrix((Nx, Nx))
    dyb = sp.csr_matrix(sp.diags([1, -1], [0, -1], shape=(Ny, Ny)))
    if pmc:
        dyb[0, 0] = 2.0
    else:
        dyb[0, 0] = 0.0
    dyb = sp.diags(1 / dls).dot(dyb)
    dyb = sp.kron(sp.eye(Nx), dyb)
    return dyb


def create_d_matrices(shape, dls, dmin_pmc=(False, False)):
    """Make the derivative matrices without PML. If dmin_pmc is True, the
    'backward' derivative in that dimension will be set to implement PMC
    boundary, otherwise it will be set to PEC."""

    dlf, dlb = dls
    dxf = make_dxf(dlf[0], shape, dmin_pmc[0])
    dxb = make_dxb(dlb[0], shape, dmin_pmc[0])
    dyf = make_dyf(dlf[1], shape, dmin_pmc[1])
    dyb = make_dyb(dlb[1], shape, dmin_pmc[1])

    return (dxf, dxb, dyf, dyb)


def create_s_matrices(omega, shape, npml, dls, eps_tensor, mu_tensor, dmin_pml=(True, True)):
    """Makes the 'S-matrices'. When dotted with derivative matrices, they add
    PML. If dmin_pml is set to False, PML will not be applied on the "bottom"
    side of the domain."""

    # strip out some information needed
    Nx, Ny = shape
    N = Nx * Ny
    nx_pml, ny_pml = npml
    # forward and backward (primal and dual) grid steps
    dlf, dlb = dls

    # Average speed (relative to C_0) in directions (xminus, xplus, yminus, yplus)
    avg_speed = average_relative_speed(Nx, Ny, npml, eps_tensor, mu_tensor)

    # Create the sfactor in each direction and for 'f' and 'b'
    s_vector_x_f = create_sfactor("f", omega, dlf[0], Nx, nx_pml, dmin_pml[0], avg_speed[:2])
    s_vector_x_b = create_sfactor("b", omega, dlb[0], Nx, nx_pml, dmin_pml[0], avg_speed[:2])
    s_vector_y_f = create_sfactor("f", omega, dlf[1], Ny, ny_pml, dmin_pml[1], avg_speed[2:])
    s_vector_y_b = create_sfactor("b", omega, dlb[1], Ny, ny_pml, dmin_pml[1], avg_speed[2:])

    # Fill the 2d space with layers of appropriate s-factors
    sx_f_2d = np.zeros(shape, dtype=np.complex128)
    sx_b_2d = np.zeros(shape, dtype=np.complex128)
    sy_f_2d = np.zeros(shape, dtype=np.complex128)
    sy_b_2d = np.zeros(shape, dtype=np.complex128)

    # Insert the cross sections into the S-grids (could be done more elegantly)
    for i in range(Ny):
        sx_f_2d[:, i] = 1 / s_vector_x_f
        sx_b_2d[:, i] = 1 / s_vector_x_b
    for i in range(Nx):
        sy_f_2d[i, :] = 1 / s_vector_y_f
        sy_b_2d[i, :] = 1 / s_vector_y_b

    # Reshape the 2d s-factors into a 1D s-vecay
    sx_f_vec = sx_f_2d.flatten()
    sx_b_vec = sx_b_2d.flatten()
    sy_f_vec = sy_f_2d.flatten()
    sy_b_vec = sy_b_2d.flatten()

    # Construct the 1D total s-vector into a diagonal matrix
    sx_f = sp.spdiags(sx_f_vec, 0, N, N)
    sx_b = sp.spdiags(sx_b_vec, 0, N, N)
    sy_f = sp.spdiags(sy_f_vec, 0, N, N)
    sy_b = sp.spdiags(sy_b_vec, 0, N, N)

    return sx_f, sx_b, sy_f, sy_b


def average_relative_speed(Nx, Ny, npml, eps_tensor, mu_tensor):
    """Compute the relative speed of light in the four pml regions by averaging the diagonal
    elements of the relative epsilon and mu within the pml region."""

    def relative_mean(tensor):
        """Mean for relative parameters. If an empty array just return 1."""
        if tensor.size == 0:
            return 1.0
        return np.mean(tensor)

    def pml_average_allsides(tensor):
        """Average ``tensor`` in the PML regions on all four sides. Returns the average values in
        order (xminus, xplus, yminus, yplus)."""

        # convert to shape (3, N) and then to (3, Nx, Ny)
        tensor_diag = np.stack([tensor[i, i, :] for i in range(3)])
        tensor_diag = tensor_diag.reshape((3, Nx, Ny))

        avg_xminus = relative_mean(tensor_diag[:, : npml[0], :])
        avg_xplus = relative_mean(tensor_diag[:, Nx - npml[0] + 1 :, :])
        avg_yminus = relative_mean(tensor_diag[:, :, : npml[1]])
        avg_yplus = relative_mean(tensor_diag[:, :, Ny - npml[1] + 1 :])
        return np.array([avg_xminus, avg_xplus, avg_yminus, avg_yplus])

    eps_avg = pml_average_allsides(eps_tensor)
    mu_avg = pml_average_allsides(mu_tensor)
    return 1 / np.sqrt(eps_avg * mu_avg)


def create_sfactor(direction, omega, dls, N, n_pml, dmin_pml, avg_speed):
    """Creates the S-factor cross section needed in the S-matrices"""

    # For no PNL, this should just be identity matrix.
    if n_pml == 0:
        return np.ones(N, dtype=np.complex128)

    # Otherwise, get different profiles for forward and reverse derivatives.
    if direction == "f":
        return create_sfactor_f(omega, dls, N, n_pml, dmin_pml, avg_speed)
    if direction == "b":
        return create_sfactor_b(omega, dls, N, n_pml, dmin_pml, avg_speed)

    raise ValueError(f"Direction value {direction} not recognized")


def create_sfactor_f(omega, dls, N, n_pml, dmin_pml, avg_speed=(1, 1)):
    """S-factor profile applied after forward derivative matrix, i.e. applied to H-field
    locations."""
    sfactor_array = np.ones(N, dtype=np.complex128)
    for i in range(N):
        if i <= n_pml - 1 and dmin_pml:
            sfactor_array[i] = s_value(dls[0], (n_pml - i - 0.5) / n_pml, omega, avg_speed[0])
        elif i >= N - n_pml:
            sfactor_array[i] = s_value(
                dls[-1], (i - (N - n_pml) + 0.5) / n_pml, omega, avg_speed[1]
            )
    return sfactor_array


def create_sfactor_b(omega, dls, N, n_pml, dmin_pml, avg_speed=(1, 1)):
    """S-factor profile applied after backward derivative matrix, i.e. applied to E-field
    locations."""
    sfactor_array = np.ones(N, dtype=np.complex128)
    for i in range(N):
        if i < n_pml and dmin_pml:
            sfactor_array[i] = s_value(dls[0], (n_pml - i) / n_pml, omega, avg_speed[0])
        elif i > N - n_pml:
            sfactor_array[i] = s_value(dls[-1], (i - (N - n_pml)) / n_pml, omega, avg_speed[1])
    return sfactor_array


def s_value(
    dl: float,
    step: int,
    omega: float,
    avg_speed: float,
    sigma_max: float = 2,
    kappa_min: float = 1,
    kappa_max: float = 3,
    order: int = 3,
):
    """S-value to use in the S-matrices.
    We use coordinate stretching formulation such that
        s(x) = kappa(x) + 1j * sigma(x) / (omega * EPSILON_0)
        kappa(x) = kappa_min + (kappa_max - kappa_min) * (x / d) ** order
        sigma(x) = sigma_max * (x / d) ** order
    where x is the position along the PML assumed to begin at 0; d is the total PML thickness;
    order is the polynomial order of the PML profile; and sigma_max is in units of
    ``avg_speed / eta_0 / dl``, with dl the mesh step and ``avg_speed`` the averaged
    speed of the wave relative to C_0 in the PML regions.

    Here x / d is given by ``step``, which has a (half-)integer value depending on whether we're
    doing forward or backward derivatives (E or H field locations along the PML normal direction).

    TODO: expose the parameters to the user.
    """

    if not sigma_max:
        # Old default value
        sigma_max = 0.8 * (order + 1)

    kappa = kappa_min + (kappa_max - kappa_min) * step**order
    sigma = sigma_max * avg_speed / (ETA_0 * dl) * step**order
    return kappa + 1j * sigma / (omega * EPSILON_0)



"""Storing tidy3d data at it's most fundamental level as xr.DataArray objects"""

from __future__ import annotations

from abc import ABC
from typing import Any, Dict, List, Mapping, Union

import autograd.numpy as anp
import dask
import h5py
import numpy as np
import pandas
import xarray as xr
from autograd.tracer import Box, getval, isbox
from xarray.core.types import InterpOptions
from xarray.core.utils import either_dict_or_kwargs

from ...constants import (
    HERTZ,
    MICROMETER,
    PICOSECOND_PER_NANOMETER_PER_KILOMETER,
    RADIAN,
    SECOND,
    WATT,
)
from ...exceptions import DataError, FileError
from ..autograd.functions import interpn
from ..types import Axis, Bound

# maps the dimension names to their attributes
DIM_ATTRS = {
    "x": {"units": MICROMETER, "long_name": "x position"},
    "y": {"units": MICROMETER, "long_name": "y position"},
    "z": {"units": MICROMETER, "long_name": "z position"},
    "f": {"units": HERTZ, "long_name": "frequency"},
    "t": {"units": SECOND, "long_name": "time"},
    "direction": {"long_name": "propagation direction"},
    "mode_index": {"long_name": "mode index"},
    "eme_port_index": {"long_name": "EME port index"},
    "eme_cell_index": {"long_name": "EME cell index"},
    "mode_index_in": {"long_name": "mode index in"},
    "mode_index_out": {"long_name": "mode index out"},
    "sweep_index": {"long_name": "sweep index"},
    "theta": {"units": RADIAN, "long_name": "elevation angle"},
    "phi": {"units": RADIAN, "long_name": "azimuth angle"},
    "ux": {"long_name": "normalized kx"},
    "uy": {"long_name": "normalized ky"},
    "orders_x": {"long_name": "diffraction order"},
    "orders_y": {"long_name": "diffraction order"},
    "face_index": {"long_name": "face index"},
    "vertex_index": {"long_name": "vertex index"},
    "axis": {"long_name": "axis"},
}


# name of the DataArray.values in the hdf5 file (xarray's default name too)
DATA_ARRAY_VALUE_NAME = "__xarray_dataarray_variable__"

# name for the autograd-traced part of the DataArray
AUTOGRAD_KEY = "AUTOGRAD"


class DataArray(xr.DataArray):
    """Subclass of ``xr.DataArray`` that requires _dims to match the keys of the coords."""

    # Always set __slots__ = () to avoid xarray warnings
    __slots__ = ()
    # stores an ordered tuple of strings corresponding to the data dimensions
    _dims = ()
    # stores a dictionary of attributes corresponding to the data values
    _data_attrs: Dict[str, str] = {}

    def __init__(self, data, *args, **kwargs):
        """Initialize ``DataArray``."""

        # initialize with untraced data
        super().__init__(getval(data), *args, **kwargs)
        # and put tracers in .attrs
        if isbox(data):
            self.attrs[AUTOGRAD_KEY] = data

    @property
    def tracers(self) -> Box:
        if self.data.size == 0:
            return None
        elif AUTOGRAD_KEY not in self.attrs and not isbox(self.data.flat[0]):
            # no tracers
            return None
        elif isbox(self.data.flat[0]):  # traced values take precedence over traced attrs
            return anp.array(self.values.tolist())
        else:
            return self.attrs[AUTOGRAD_KEY]

    @classmethod
    def __get_validators__(cls):
        """Validators that get run when :class:`.DataArray` objects are added to pydantic models."""
        yield cls.check_unloaded_data
        yield cls.validate_dims
        yield cls.assign_data_attrs
        yield cls.assign_coord_attrs

    @classmethod
    def check_unloaded_data(cls, val):
        """If the data comes in as the raw data array string, raise a custom warning."""
        if isinstance(val, str) and val in DATA_ARRAY_MAP:
            raise DataError(
                f"Trying to load {cls.__name__} but the data is not present. "
                "Note that data will not be saved to .json file. "
                "use .hdf5 format instead if data present."
            )
        return cls(val)

    @classmethod
    def validate_dims(cls, val):
        """Make sure the dims are the same as _dims, then put them in the correct order."""
        if set(val.dims) != set(cls._dims):
            raise ValueError(f"wrong dims, expected '{cls._dims}', got '{val.dims}'")
        return val.transpose(*cls._dims)

    @classmethod
    def assign_data_attrs(cls, val):
        """Assign the correct data attributes to the :class:`.DataArray`."""

        for attr_name, attr in cls._data_attrs.items():
            val.attrs[attr_name] = attr
        return val

    def _interp_validator(self, field_name: str = None) -> None:
        """Make sure we can interp()/sel() the data.

        NOTE
        ----
        This does not check every 'DataArray' by default. Instead, when required, this check can be
        called from a validator, as is the case with 'CustomMedium' and 'CustomFieldSource'.
        """
        # skip this validator if currently tracing for autograd because
        # self.values will be dtype('object') and not interpolatable
        if self.tracers is not None:
            return

        if field_name is None:
            field_name = "DataArray"

        dims = self.coords.dims

        for dim in dims:
            # in case we encounter some /0 or /NaN we'll ignore the warnings here
            with np.errstate(divide="ignore", invalid="ignore"):
                # check that we can interpolate
                try:
                    x0 = np.array(self.coords[dim][0])
                    self.interp({dim: x0}, method="linear")
                    self.interp({dim: x0}, method="nearest")
                    # self.interp_like(self.isel({self.dim: 0}))
                except pandas.errors.InvalidIndexError as e:
                    raise DataError(
                        f"'{field_name}.interp()' fails to interpolate along {dim} which is used by the solver. "
                        "This may be caused, for instance, by duplicated data "
                        f"in this dimension (you can verify this by running "
                        f"'{field_name}={field_name}.drop_duplicates(dim=\"{dim}\")' "
                        f"and interpolate with the new '{field_name}'). "
                        "Plase make sure data can be interpolated."
                    ) from e
                # in case it can interpolate, try also to sel
                try:
                    x0 = np.array(self.coords[dim][0])
                    self.sel({dim: x0}, method="nearest")
                except pandas.errors.InvalidIndexError as e:
                    raise DataError(
                        f"'{field_name}.sel()' fails to select along {dim} which is used by the solver. "
                        "This may be caused, for instance, by duplicated data "
                        f"in this dimension (you can verify this by running "
                        f"'{field_name}={field_name}.drop_duplicates(dim=\"{dim}\")' "
                        f"and run 'sel()' with the new '{field_name}'). "
                        "Plase make sure 'sel()' can be used on the 'DataArray'."
                    ) from e

    @classmethod
    def assign_coord_attrs(cls, val):
        """Assign the correct coordinate attributes to the :class:`.DataArray`."""

        for dim in cls._dims:
            dim_attrs = DIM_ATTRS.get(dim)
            if dim_attrs is not None:
                for attr_name, attr in dim_attrs.items():
                    val.coords[dim].attrs[attr_name] = attr
        return val

    @classmethod
    def __modify_schema__(cls, field_schema):
        """Sets the schema of DataArray object."""

        schema = dict(
            title="DataArray",
            type="xr.DataArray",
            properties=dict(
                _dims=dict(
                    title="_dims",
                    type="Tuple[str, ...]",
                ),
            ),
            required=["_dims"],
        )
        field_schema.update(schema)

    @classmethod
    def _json_encoder(cls, val):
        """What function to call when writing a DataArray to json."""
        return type(val).__name__

    def __eq__(self, other) -> bool:
        """Whether two data array objects are equal."""

        if not isinstance(other, xr.DataArray):
            return False

        if not self.data.shape == other.data.shape or not np.all(self.data == other.data):
            return False
        for key, val in self.coords.items():
            if not np.all(np.array(val) == np.array(other.coords[key])):
                return False
        return True

    @property
    def abs(self):
        """Absolute value of data array."""
        return abs(self)

    @property
    def is_uniform(self):
        """Whether each element is of equal value in the data array"""
        raw_data = self.data.ravel()
        return np.allclose(raw_data, raw_data[0])

    def to_hdf5(self, fname: Union[str, h5py.File], group_path: str) -> None:
        """Save an xr.DataArray to the hdf5 file or file handle with a given path to the group."""

        # file name passed
        if isinstance(fname, str):
            with h5py.File(fname, "w") as f_handle:
                self.to_hdf5_handle(f_handle=f_handle, group_path=group_path)

        # file handle passed
        else:
            self.to_hdf5_handle(f_handle=fname, group_path=group_path)

    def to_hdf5_handle(self, f_handle: h5py.File, group_path: str) -> None:
        """Save an xr.DataArray to the hdf5 file handle with a given path to the group."""

        sub_group = f_handle.create_group(group_path)
        sub_group[DATA_ARRAY_VALUE_NAME] = self.values
        for key, val in self.coords.items():
            if val.dtype == "<U1":
                sub_group[key] = val.values.tolist()
            else:
                sub_group[key] = val

    @classmethod
    def from_hdf5(cls, fname: str, group_path: str) -> DataArray:
        """Load an DataArray from an hdf5 file with a given path to the group."""
        with h5py.File(fname, "r") as f:
            sub_group = f[group_path]
            values = np.array(sub_group[DATA_ARRAY_VALUE_NAME])
            coords = {dim: np.array(sub_group[dim]) for dim in cls._dims if dim in sub_group}
            for key, val in coords.items():
                if val.dtype == "O":
                    coords[key] = [byte_string.decode() for byte_string in val.tolist()]
            return cls(values, coords=coords, dims=cls._dims)

    @classmethod
    def from_file(cls, fname: str, group_path: str) -> DataArray:
        """Load an DataArray from an hdf5 file with a given path to the group."""
        if ".hdf5" not in fname:
            raise FileError(
                "DataArray objects must be written to '.hdf5' format. "
                f"Given filename of {fname}."
            )
        return cls.from_hdf5(fname=fname, group_path=group_path)

    def __hash__(self) -> int:
        """Generate hash value for a :class:.`DataArray` instance, needed for custom components."""
        token_str = dask.base.tokenize(self)
        return hash(token_str)

    def multiply_at(self, value: complex, coord_name: str, indices: List[int]) -> DataArray:
        """Multiply self by value at indices into ."""
        self_mult = self.copy()
        self_mult[{coord_name: indices}] *= value
        return self_mult

    def interp(
        self,
        coords: Union[Mapping[Any, Any], None] = None,
        method: InterpOptions = "linear",
        assume_sorted: bool = False,
        kwargs: Union[Mapping[str, Any], None] = None,
        **coords_kwargs: Any,
    ):
        """Interpolate this DataArray to new coordinate values.

        Parameters
        ----------
        coords : Union[Mapping[Any, Any], None] = None
            A mapping from dimension names to new coordinate labels.
        method : InterpOptions = "linear"
            The interpolation method to use.
        assume_sorted : bool = False
            If True, skip sorting of coordinates.
        kwargs : Union[Mapping[str, Any], None] = None
            Additional keyword arguments to pass to the interpolation function.
        **coords_kwargs : Any
            The keyword arguments form of coords.

        Returns
        -------
        DataArray
            A new DataArray with interpolated values.

        Raises
        ------
        KeyError
            If any of the specified coordinates are not in the DataArray.
        """
        if self.tracers is not None:  # use custom interp if using traced data
            coords = either_dict_or_kwargs(coords, coords_kwargs, "interp")

            missing_keys = set(coords) - set(self.coords)
            if missing_keys:
                raise KeyError(f"Cannot interpolate: {missing_keys} not in coords.")

            obj = self if assume_sorted else self.sortby(list(coords.keys()))

            out_coords = {k: coords.get(k, obj.coords[k]) for k in obj.dims}
            points = tuple(obj.coords[k] for k in obj.dims)
            xi = tuple(out_coords.values())

            vals = interpn(points, obj.tracers, xi, method=method)

            da = DataArray(vals, out_coords)  # tracers go into .attrs
            if isbox(self.values.flat[0]):  # if tracing .values instead of .attrs
                da = da.copy(deep=False, data=vals)  # copy over tracers

            return da

        return super().interp(
            coords=coords,
            method=method,
            assume_sorted=assume_sorted,
            kwargs=kwargs,
            **coords_kwargs,
        )

    def conj(self, *args: Any, **kwargs: Any):
        """Return the complex conjugate of this DataArray."""
        if self.tracers is not None:
            return self.__array_wrap__(anp.conj(self.tracers))
        return super().conj(*args, **kwargs)

    @property
    def real(self):
        """Return the real part of this DataArray."""
        if self.tracers is not None:
            return self.__array_wrap__(anp.real(self.tracers))
        return super().real


class FreqDataArray(DataArray):
    """Frequency-domain array.

    Example
    -------
    >>> f = [2e14, 3e14]
    >>> fd = FreqDataArray((1+1j) * np.random.random((2,)), coords=dict(f=f))
    """

    __slots__ = ()
    _dims = ("f",)


class FreqModeDataArray(DataArray):
    """Array over frequency and mode index.

    Example
    -------
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(5)
    >>> coords = dict(f=f, mode_index=mode_index)
    >>> fd = FreqModeDataArray((1+1j) * np.random.random((2, 5)), coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "mode_index")


class TimeDataArray(DataArray):
    """Time-domain array.

    Example
    -------
    >>> t = [0, 1e-12, 2e-12]
    >>> td = TimeDataArray((1+1j) * np.random.random((3,)), coords=dict(t=t))
    """

    __slots__ = ()
    _dims = "t"


class MixedModeDataArray(DataArray):
    """Scalar property associated with mode pairs

    Example
    -------
    >>> f = [1e14, 2e14, 3e14]
    >>> mode_index_0 = np.arange(4)
    >>> mode_index_1 = np.arange(2)
    >>> coords = dict(f=f, mode_index_0=mode_index_0, mode_index_1=mode_index_1)
    >>> data = MixedModeDataArray((1+1j) * np.random.random((3, 4, 2)), coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "mode_index_0", "mode_index_1")


class AbstractSpatialDataArray(DataArray, ABC):
    """Spatial distribution."""

    __slots__ = ()
    _dims = ("x", "y", "z")
    _data_attrs = {"long_name": "field value"}

    @property
    def _spatially_sorted(self) -> SpatialDataArray:
        """Check whether sorted and sort if not."""
        needs_sorting = []
        for axis in "xyz":
            axis_coords = self.coords[axis].values
            if len(axis_coords) > 1 and np.any(axis_coords[1:] < axis_coords[:-1]):
                needs_sorting.append(axis)

        if len(needs_sorting) > 0:
            return self.sortby(needs_sorting)

        return self

    def sel_inside(self, bounds: Bound) -> SpatialDataArray:
        """Return a new SpatialDataArray that contains the minimal amount data necessary to cover
        a spatial region defined by ``bounds``. Note that the returned data is sorted with respect
        to spatial coordinates.


        Parameters
        ----------
        bounds : Tuple[float, float, float], Tuple[float, float float]
            Min and max bounds packaged as ``(minx, miny, minz), (maxx, maxy, maxz)``.

        Returns
        -------
        SpatialDataArray
            Extracted spatial data array.
        """
        if any(bmin > bmax for bmin, bmax in zip(*bounds)):
            raise DataError(
                "Min and max bounds must be packaged as '(minx, miny, minz), (maxx, maxy, maxz)'."
            )

        # make sure data is sorted with respect to coordinates
        sorted_self = self._spatially_sorted

        inds_list = []

        coords = (sorted_self.x, sorted_self.y, sorted_self.z)

        for coord, smin, smax in zip(coords, bounds[0], bounds[1]):
            length = len(coord)

            # one point along direction, assume invariance
            if length == 1:
                comp_inds = [0]
            else:
                # if data does not cover structure at all take the closest index
                if smax < coord[0]:  # structure is completely on the left side
                    # take 2 if possible, so that linear iterpolation is possible
                    comp_inds = np.arange(0, max(2, length))

                elif smin > coord[-1]:  # structure is completely on the right side
                    # take 2 if possible, so that linear iterpolation is possible
                    comp_inds = np.arange(min(0, length - 2), length)

                else:
                    if smin < coord[0]:
                        ind_min = 0
                    else:
                        ind_min = max(0, (coord >= smin).argmax().data - 1)

                    if smax > coord[-1]:
                        ind_max = length - 1
                    else:
                        ind_max = (coord >= smax).argmax().data

                    comp_inds = np.arange(ind_min, ind_max + 1)

            inds_list.append(comp_inds)

        return sorted_self.isel(x=inds_list[0], y=inds_list[1], z=inds_list[2])

    def does_cover(self, bounds: Bound) -> bool:
        """Check whether data fully covers specified by ``bounds`` spatial region. If data contains
        only one point along a given direction, then it is assumed the data is constant along that
        direction and coverage is not checked.


        Parameters
        ----------
        bounds : Tuple[float, float, float], Tuple[float, float float]
            Min and max bounds packaged as ``(minx, miny, minz), (maxx, maxy, maxz)``.

        Returns
        -------
        bool
            Full cover check outcome.
        """
        if any(bmin > bmax for bmin, bmax in zip(*bounds)):
            raise DataError(
                "Min and max bounds must be packaged as '(minx, miny, minz), (maxx, maxy, maxz)'."
            )

        coords = (self.x, self.y, self.z)
        return all(
            (np.min(coord) <= smin and np.max(coord) >= smax) or len(coord) == 1
            for coord, smin, smax in zip(coords, bounds[0], bounds[1])
        )


class SpatialDataArray(AbstractSpatialDataArray):
    """Spatial distribution.

    Example
    -------
    >>> x = [1,2]
    >>> y = [2,3,4]
    >>> z = [3,4,5,6]
    >>> coords = dict(x=x, y=y, z=z)
    >>> fd = SpatialDataArray((1+1j) * np.random.random((2,3,4)), coords=coords)
    """

    __slots__ = ()

    def reflect(self, axis: Axis, center: float, reflection_only: bool = False) -> SpatialDataArray:
        """Reflect data across the plane define by parameters ``axis`` and ``center`` from right to
        left. Note that the returned data is sorted with respect to spatial coordinates.

        Parameters
        ----------
        axis : Literal[0, 1, 2]
            Normal direction of the reflection plane.
        center : float
            Location of the reflection plane along its normal direction.
        reflection_only : bool = False
            Return only reflected data.

        Returns
        -------
        SpatialDataArray
            Data after reflection is performed.
        """

        sorted_self = self._spatially_sorted

        coords = [sorted_self.x.values, sorted_self.y.values, sorted_self.z.values]
        data = np.array(sorted_self.data)

        data_left_bound = coords[axis][0]

        if np.isclose(center, data_left_bound):
            num_duplicates = 1
        elif center > data_left_bound:
            raise DataError("Reflection center must be outside and to the left of the data region.")
        else:
            num_duplicates = 0

        if reflection_only:
            coords[axis] = 2 * center - coords[axis]
            coords_dict = dict(zip("xyz", coords))

            tmp_arr = SpatialDataArray(sorted_self.data, coords=coords_dict)

            return tmp_arr.sortby("xyz"[axis])

        shape = np.array(np.shape(data))
        old_len = shape[axis]
        shape[axis] = 2 * old_len - num_duplicates

        ind_left = [slice(shape[0]), slice(shape[1]), slice(shape[2])]
        ind_right = [slice(shape[0]), slice(shape[1]), slice(shape[2])]

        ind_left[axis] = slice(old_len - 1, None, -1)
        ind_right[axis] = slice(old_len - num_duplicates, None)

        new_data = np.zeros(shape)

        new_data[ind_left[0], ind_left[1], ind_left[2]] = data
        new_data[ind_right[0], ind_right[1], ind_right[2]] = data

        new_coords = np.zeros(shape[axis])
        new_coords[old_len - num_duplicates :] = coords[axis]
        new_coords[old_len - 1 :: -1] = 2 * center - coords[axis]

        coords[axis] = new_coords
        coords_dict = dict(zip("xyz", coords))

        return SpatialDataArray(new_data, coords=coords_dict)


class ScalarFieldDataArray(AbstractSpatialDataArray):
    """Spatial distribution in the frequency-domain.

    Example
    -------
    >>> x = [1,2]
    >>> y = [2,3,4]
    >>> z = [3,4,5,6]
    >>> f = [2e14, 3e14]
    >>> coords = dict(x=x, y=y, z=z, f=f)
    >>> fd = ScalarFieldDataArray((1+1j) * np.random.random((2,3,4,2)), coords=coords)
    """

    __slots__ = ()
    _dims = ("x", "y", "z", "f")


class ScalarFieldTimeDataArray(AbstractSpatialDataArray):
    """Spatial distribution in the time-domain.

    Example
    -------
    >>> x = [1,2]
    >>> y = [2,3,4]
    >>> z = [3,4,5,6]
    >>> t = [0, 1e-12, 2e-12]
    >>> coords = dict(x=x, y=y, z=z, t=t)
    >>> fd = ScalarFieldTimeDataArray(np.random.random((2,3,4,3)), coords=coords)
    """

    __slots__ = ()
    _dims = ("x", "y", "z", "t")


class ScalarModeFieldDataArray(AbstractSpatialDataArray):
    """Spatial distribution of a mode in frequency-domain as a function of mode index.

    Example
    -------
    >>> x = [1,2]
    >>> y = [2,3,4]
    >>> z = [3,4,5,6]
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(5)
    >>> coords = dict(x=x, y=y, z=z, f=f, mode_index=mode_index)
    >>> fd = ScalarModeFieldDataArray((1+1j) * np.random.random((2,3,4,2,5)), coords=coords)
    """

    __slots__ = ()
    _dims = ("x", "y", "z", "f", "mode_index")


class FluxDataArray(DataArray):
    """Flux through a surface in the frequency-domain.

    Example
    -------
    >>> f = [2e14, 3e14]
    >>> coords = dict(f=f)
    >>> fd = FluxDataArray(np.random.random(2), coords=coords)
    """

    __slots__ = ()
    _dims = ("f",)
    _data_attrs = {"units": WATT, "long_name": "flux"}


class FluxTimeDataArray(DataArray):
    """Flux through a surface in the time-domain.

    Example
    -------
    >>> t = [0, 1e-12, 2e-12]
    >>> coords = dict(t=t)
    >>> data = FluxTimeDataArray(np.random.random(3), coords=coords)
    """

    __slots__ = ()
    _dims = ("t",)
    _data_attrs = {"units": WATT, "long_name": "flux"}


class ModeAmpsDataArray(DataArray):
    """Forward and backward propagating complex-valued mode amplitudes.

    Example
    -------
    >>> direction = ["+", "-"]
    >>> f = [1e14, 2e14, 3e14]
    >>> mode_index = np.arange(4)
    >>> coords = dict(direction=direction, f=f, mode_index=mode_index)
    >>> data = ModeAmpsDataArray((1+1j) * np.random.random((2, 3, 4)), coords=coords)
    """

    __slots__ = ()
    _dims = ("direction", "f", "mode_index")
    _data_attrs = {"units": "sqrt(W)", "long_name": "mode amplitudes"}


class ModeIndexDataArray(DataArray):
    """Complex-valued effective propagation index of a mode.

    Example
    -------
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(4)
    >>> coords = dict(f=f, mode_index=mode_index)
    >>> data = ModeIndexDataArray((1+1j) * np.random.random((2,4)), coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "mode_index")
    _data_attrs = {"long_name": "Propagation index"}


class GroupIndexDataArray(DataArray):
    """Group index of a mode.

    Example
    -------
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(4)
    >>> coords = dict(f=f, mode_index=mode_index)
    >>> data = GroupIndexDataArray((1+1j) * np.random.random((2,4)), coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "mode_index")
    _data_attrs = {"long_name": "Group index"}


class ModeDispersionDataArray(DataArray):
    """Dispersion parameter of a mode.

    Example
    -------
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(4)
    >>> coords = dict(f=f, mode_index=mode_index)
    >>> data = ModeDispersionDataArray((1+1j) * np.random.random((2,4)), coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "mode_index")
    _data_attrs = {
        "long_name": "Dispersion parameter",
        "units": PICOSECOND_PER_NANOMETER_PER_KILOMETER,
    }


class FieldProjectionAngleDataArray(DataArray):
    """Far fields in frequency domain as a function of angles theta and phi.

    Example
    -------
    >>> f = np.linspace(1e14, 2e14, 10)
    >>> r = np.atleast_1d(5)
    >>> theta = np.linspace(0, np.pi, 10)
    >>> phi = np.linspace(0, 2*np.pi, 20)
    >>> coords = dict(r=r, theta=theta, phi=phi, f=f)
    >>> values = (1+1j) * np.random.random((len(r), len(theta), len(phi), len(f)))
    >>> data = FieldProjectionAngleDataArray(values, coords=coords)
    """

    __slots__ = ()
    _dims = ("r", "theta", "phi", "f")
    _data_attrs = {"long_name": "radiation vectors"}


class FieldProjectionCartesianDataArray(DataArray):
    """Far fields in frequency domain as a function of local x and y coordinates.

    Example
    -------
    >>> f = np.linspace(1e14, 2e14, 10)
    >>> x = np.linspace(0, 5, 10)
    >>> y = np.linspace(0, 10, 20)
    >>> z = np.atleast_1d(5)
    >>> coords = dict(x=x, y=y, z=z, f=f)
    >>> values = (1+1j) * np.random.random((len(x), len(y), len(z), len(f)))
    >>> data = FieldProjectionCartesianDataArray(values, coords=coords)
    """

    __slots__ = ()
    _dims = ("x", "y", "z", "f")
    _data_attrs = {"long_name": "radiation vectors"}


class FieldProjectionKSpaceDataArray(DataArray):
    """Far fields in frequency domain as a function of normalized
    kx and ky vectors on the observation plane.

    Example
    -------
    >>> f = np.linspace(1e14, 2e14, 10)
    >>> r = np.atleast_1d(5)
    >>> ux = np.linspace(0, 5, 10)
    >>> uy = np.linspace(0, 10, 20)
    >>> coords = dict(ux=ux, uy=uy, r=r, f=f)
    >>> values = (1+1j) * np.random.random((len(ux), len(uy), len(r), len(f)))
    >>> data = FieldProjectionKSpaceDataArray(values, coords=coords)
    """

    __slots__ = ()
    _dims = ("ux", "uy", "r", "f")
    _data_attrs = {"long_name": "radiation vectors"}


class DiffractionDataArray(DataArray):
    """Diffraction power amplitudes as a function of diffraction orders and frequency.

    Example
    -------
    >>> f = np.linspace(1e14, 2e14, 10)
    >>> orders_x = np.linspace(-1, 1, 3)
    >>> orders_y = np.linspace(-2, 2, 5)
    >>> coords = dict(orders_x=orders_x, orders_y=orders_y, f=f)
    >>> values = (1+1j) * np.random.random((len(orders_x), len(orders_y), len(f)))
    >>> data = DiffractionDataArray(values, coords=coords)
    """

    __slots__ = ()
    _dims = ("orders_x", "orders_y", "f")
    _data_attrs = {"long_name": "diffraction amplitude"}


class TriangleMeshDataArray(DataArray):
    """Data of the triangles of a surface mesh as in the STL file format."""

    __slots__ = ()
    _dims = ("face_index", "vertex_index", "axis")
    _data_attrs = {"long_name": "surface mesh triangles"}


class HeatDataArray(DataArray):
    """Heat data array.

    Example
    -------
    >>> T = [0, 1e-12, 2e-12]
    >>> td = HeatDataArray((1+1j) * np.random.random((3,)), coords=dict(T=T))
    """

    __slots__ = ()
    _dims = "T"


class EMEScalarModeFieldDataArray(AbstractSpatialDataArray):
    """Spatial distribution of a mode in frequency-domain as a function of mode index
    and EME cell index.

    Example
    -------
    >>> x = [1,2]
    >>> y = [2,3,4]
    >>> z = [3]
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(5)
    >>> eme_cell_index = np.arange(5)
    >>> coords = dict(x=x, y=y, z=z, f=f, mode_index=mode_index, eme_cell_index=eme_cell_index)
    >>> fd = EMEScalarModeFieldDataArray((1+1j) * np.random.random((2,3,1,2,5,5)), coords=coords)
    """

    __slots__ = ()
    _dims = ("x", "y", "z", "f", "sweep_index", "eme_cell_index", "mode_index")


class EMEFreqModeDataArray(DataArray):
    """Array over frequency, mode index, and EME cell index.

    Example
    -------
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(5)
    >>> eme_cell_index = np.arange(5)
    >>> coords = dict(f=f, mode_index=mode_index, eme_cell_index=eme_cell_index)
    >>> fd = EMEFreqModeDataArray((1+1j) * np.random.random((2, 5, 5)), coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "sweep_index", "eme_cell_index", "mode_index")


class EMEScalarFieldDataArray(AbstractSpatialDataArray):
    """Spatial distribution of a field excited from an EME port in frequency-domain as a
    function of mode index at the EME port and the EME port index.

    Example
    -------
    >>> x = [1,2]
    >>> y = [2,3,4]
    >>> z = [3,4,5,6]
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(5)
    >>> eme_port_index = [0, 1]
    >>> coords = dict(x=x, y=y, z=z, f=f, mode_index=mode_index, eme_port_index=eme_port_index)
    >>> fd = EMEScalarFieldDataArray((1+1j) * np.random.random((2,3,4,2,5,2)), coords=coords)
    """

    __slots__ = ()
    _dims = ("x", "y", "z", "f", "sweep_index", "eme_port_index", "mode_index")


class EMECoefficientDataArray(DataArray):
    """EME expansion coefficient of the mode `mode_index_out` in the EME cell
    `eme_cell_index`, when excited from mode `mode_index_in` of EME port `eme_port_index`.

    Example
    -------
    >>> mode_index_in = [0, 1]
    >>> mode_index_out = [0, 1]
    >>> eme_cell_index = np.arange(5)
    >>> eme_port_index = [0, 1]
    >>> f = [2e14]
    >>> coords = dict(
    ...     f=f,
    ...     mode_index_out=mode_index_out,
    ...     mode_index_in=mode_index_in,
    ...     eme_cell_index=eme_cell_index,
    ...     eme_port_index=eme_port_index
    ... )
    >>> fd = EMECoefficientDataArray((1 + 1j) * np.random.random((1, 2, 2, 5, 2)), coords=coords)
    """

    __slots__ = ()
    _dims = (
        "f",
        "sweep_index",
        "eme_port_index",
        "eme_cell_index",
        "mode_index_out",
        "mode_index_in",
    )
    _data_attrs = {"long_name": "mode expansion coefficient"}


class EMESMatrixDataArray(DataArray):
    """Scattering matrix elements for a fixed pair of ports, possibly with an extra
    sweep index.

    Example
    -------
    >>> mode_index_in = [0, 1]
    >>> mode_index_out = [0, 1, 2]
    >>> f = [2e14]
    >>> sweep_index = np.arange(10)
    >>> coords = dict(
    ...     f=f,
    ...     mode_index_out=mode_index_out,
    ...     mode_index_in=mode_index_in,
    ...     sweep_index=sweep_index
    ... )
    >>> fd = EMESMatrixDataArray((1 + 1j) * np.random.random((1, 3, 2, 10)), coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "sweep_index", "mode_index_out", "mode_index_in")
    _data_attrs = {"long_name": "scattering matrix element"}


class EMEModeIndexDataArray(DataArray):
    """Complex-valued effective propagation index of an EME mode,
    also indexed by EME cell.

    Example
    -------
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(4)
    >>> eme_cell_index = np.arange(5)
    >>> coords = dict(f=f, mode_index=mode_index, eme_cell_index=eme_cell_index)
    >>> data = EMEModeIndexDataArray((1+1j) * np.random.random((2,4,5)), coords=coords)
    """

    __slots__ = ()
    _dims = ("f", "sweep_index", "eme_cell_index", "mode_index")
    _data_attrs = {"long_name": "Propagation index"}


class ChargeDataArray(DataArray):
    """Charge data array.

    Example
    -------
    >>> n = [0, 1e-12, 2e-12]
    >>> p = [0, 3e-12, 4e-12]
    >>> td = ChargeDataArray((1+1j) * np.random.random((3,3)), coords=dict(n=n, p=p))
    """

    __slots__ = ()
    _dims = ("n", "p")


class PointDataArray(DataArray):
    """A two-dimensional array that stores coordinates of a collection of points.
    Dimension ``index`` denotes the index of a point in the collection, and dimension ``axis``
    denotes the point's coordinate along that axis.

    Example
    -------
    >>> point_array = PointDataArray(
    ...     (1+1j) * np.random.random((5, 3)), coords=dict(index=np.arange(5), axis=np.arange(3)),
    ... )
    >>> # get coordinates of a point number 3
    >>> point3 = point_array.sel(index=3)
    >>> # get x coordinates of all points
    >>> x_coords = point_array.sel(axis=0)
    """

    __slots__ = ()
    _dims = ("index", "axis")


class CellDataArray(DataArray):
    """A two-dimensional array that stores indices of points composing each cell in a collection of
    cells of the same type (for example: triangles, tetrahedra, etc). Dimension ``cell_index``
    denotes the index of a cell in the collection, and dimension ``vertex_index`` denotes placement
    (index) of a point in a cell (for example: 0, 1, or 2 for triangles; 0, 1, 2, or 3 for
    tetrahedra).

    Example
    -------
    >>> cell_array = CellDataArray(
    ...     (1+1j) * np.random.random((4, 3)),
    ...     coords=dict(cell_index=np.arange(4), vertex_index=np.arange(3)),
    ... )
    >>> # get indices of points composing cell number 3
    >>> cell3 = cell_array.sel(cell_index=3)
    >>> # get indices of points that represent the first vertex in each cell
    >>> first_vertices = cell_array.sel(vertex_index=0)
    """

    __slots__ = ()
    _dims = ("cell_index", "vertex_index")


class IndexedDataArray(DataArray):
    """Stores a one-dimensional array enumerated by coordinate ``index``. It is typically used
    in conjuction with a ``PointDataArray`` to store point-associated data or a ``CellDataArray``
    to store cell-associated data.

    Example
    -------
    >>> indexed_array = IndexedDataArray(
    ...     (1+1j) * np.random.random((3,)), coords=dict(index=np.arange(3))
    ... )
    """

    __slots__ = ()
    _dims = ("index",)


DATA_ARRAY_TYPES = [
    SpatialDataArray,
    ScalarFieldDataArray,
    ScalarFieldTimeDataArray,
    ScalarModeFieldDataArray,
    FluxDataArray,
    FluxTimeDataArray,
    ModeAmpsDataArray,
    ModeIndexDataArray,
    GroupIndexDataArray,
    ModeDispersionDataArray,
    FieldProjectionAngleDataArray,
    FieldProjectionCartesianDataArray,
    FieldProjectionKSpaceDataArray,
    DiffractionDataArray,
    FreqModeDataArray,
    FreqDataArray,
    TimeDataArray,
    FreqModeDataArray,
    TriangleMeshDataArray,
    HeatDataArray,
    EMEScalarFieldDataArray,
    EMEScalarModeFieldDataArray,
    EMESMatrixDataArray,
    EMECoefficientDataArray,
    EMEModeIndexDataArray,
    EMEFreqModeDataArray,
    ChargeDataArray,
    PointDataArray,
    CellDataArray,
    IndexedDataArray,
]
DATA_ARRAY_MAP = {data_array.__name__: data_array for data_array in DATA_ARRAY_TYPES}



"""Defines the methods used for parameter sweep."""

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Tuple, Union

import numpy as np
import pydantic.v1 as pd
import scipy.stats.qmc as qmc

from ... import web
from ...components.base import Tidy3dBaseModel
from ...components.simulation import Simulation
from ...log import log
from ...web.api.container import BatchData
from .parameter import ParameterType

DEFAULT_MONTE_CARLO_SAMPLER_TYPE = qmc.LatinHypercube


class Method(Tidy3dBaseModel, ABC):
    """Spec for a sweep algorithm, with a method to run it."""

    name: str = pd.Field(None, title="Name", description="Optional name for the sweep method.")

    @abstractmethod
    def run(self, parameters: Tuple[ParameterType, ...], fn: Callable) -> Tuple[Any]:
        """Defines the search algorithm (sequential)."""

    @abstractmethod
    def sample(self, parameters: Tuple[ParameterType, ...], **kwargs) -> Dict[str, Any]:
        """Defines how the design parameters are sampled."""

    @staticmethod
    def assert_hashable(fn_args: dict) -> None:
        """Raise error if the function arguments aren't hashable (do before computation)."""
        fn_args_tuple = tuple(fn_args)
        try:
            hash(fn_args_tuple)
        except TypeError:
            raise ValueError(
                "Function arguments must be hashable. "
                "Parameter sweep tool won't work with sets of lists, dicts or numpy arrays. "
                "Convert these to 'tuple' for a workaround."
            )


class MethodIndependent(Method, ABC):
    """A sweep method where all points are independently computed."""

    @staticmethod
    def get_num_points(fn_args: Dict[str, tuple]) -> int:
        """Compute number of points from the function arguments and do error checking."""
        num_points_each_dim = [len(val) for val in fn_args.values()]
        if len(set(num_points_each_dim)) != 1:
            raise ValueError(
                f"Found different number of points: {num_points_each_dim} along each dimension. "
                "This suggests a bug in the parameter sweep tool. "
                "Please raise an issue on the front end GitHub repository."
            )
        return num_points_each_dim[0]

    def _assemble_args(self, parameters: Tuple[ParameterType, ...]) -> Tuple[dict, int]:
        """Sample design parameters, check the args are hashable and compute number of points."""

        fn_args = self.sample(parameters)
        self.assert_hashable(fn_args)
        num_points = self.get_num_points(fn_args)
        return fn_args, num_points

    def run(self, parameters: Tuple[ParameterType, ...], fn: Callable) -> Tuple[Any]:
        """Defines the search algorithm (sequential)."""

        # get all function inputs
        fn_args, num_points = self._assemble_args(parameters)

        # for each point, construct the function inputs, run it, record output
        result = []
        for i in range(num_points):
            fn_kwargs = {key: vals[i] for key, vals in fn_args.items()}
            fn_output = fn(**fn_kwargs)
            result.append(fn_output)

        return fn_args, result

    def _run_batch(
        self, simulations: Dict[str, Simulation], path_dir: str = None, **kwargs
    ) -> BatchData:
        """Create a batch of simulations and run it. Mainly separated out for ease of testing."""
        batch = web.Batch(simulations=simulations, simulation_type="tidy3d_design", **kwargs)

        if path_dir:
            run_kwargs = dict(path_dir=path_dir)
        else:
            run_kwargs = {}
        return batch.run(**run_kwargs)

    def run_batch(
        self,
        parameters: Tuple[ParameterType, ...],
        fn_pre: Callable,
        fn_post: Callable,
        path_dir: str = None,
        **batch_kwargs,
    ) -> Tuple[Any]:
        """Defines the search algorithm (batched)."""

        # get all function inputs
        fn_args, num_points = self._assemble_args(parameters)

        def get_task_name(pt_index: int, sim_index: int, fn_kwargs: dict) -> str:
            """Get task name for 'index'-th set of function kwargs."""
            try:
                kwarg_str = str(fn_kwargs)
                if sim_index is not None:
                    return f"{kwarg_str}_{sim_index}"
                return kwarg_str
            # just to be safe, handle the case if this does not work
            except ValueError:
                return f"{pt_index}_{sim_index}"

        # for each point, construct the simulation inputs into a dict
        simulations = {}
        task_name_mappings = []
        for i in range(num_points):
            fn_kwargs = {key: vals[i] for key, vals in fn_args.items()}
            sim = fn_pre(**fn_kwargs)
            if isinstance(sim, Simulation):
                task_name = get_task_name(pt_index=i, sim_index=None, fn_kwargs=fn_kwargs)
                simulations[task_name] = sim
                task_name_mappings.append([task_name])
            elif isinstance(sim, dict):
                task_name_mappings.append({})
                for name, _sim in sim.items():
                    task_name = get_task_name(pt_index=1, sim_index=name, fn_kwargs=fn_kwargs)
                    simulations[task_name] = _sim
                    task_name_mappings[i][name] = task_name
            else:
                task_name_mappings.append([])
                for j, _sim in enumerate(sim):
                    task_name = get_task_name(pt_index=i, sim_index=j, fn_kwargs=fn_kwargs)
                    simulations[task_name] = _sim
                    task_name_mappings[i].append(task_name)

        # run in batch
        batch_data = self._run_batch(simulations=simulations, path_dir=path_dir, **batch_kwargs)
        task_id_dict = batch_data.task_ids

        # run post processing on each data
        result = []
        task_ids = []
        for task_names_i in task_name_mappings:
            if isinstance(task_names_i, dict):
                task_ids.append([task_id_dict[task_name] for task_name in task_names_i.values()])
                kwargs_dict = {
                    kwarg_name: batch_data[task_name]
                    for kwarg_name, task_name in task_names_i.items()
                }
                val = fn_post(**kwargs_dict)
            else:
                task_ids.append([task_id_dict[task_name] for task_name in task_names_i])
                args_list = (batch_data[task_name] for task_name in task_names_i)
                val = fn_post(*args_list)

            result.append(val)

        return fn_args, result, task_ids, batch_data


class MethodGrid(MethodIndependent):
    """Select parameters uniformly on a grid.

    Example
    -------
    >>> import tidy3d.plugins.design as tdd
    >>> method = tdd.MethodGrid()
    """

    def sample(self, parameters: Tuple[ParameterType, ...], **kwargs) -> Dict[str, Any]:
        """Defines how the design parameters are sampled on grid."""

        # sample each dimension individually
        vals_each_dim = {}
        for design_var in parameters:
            vals = design_var.sample_grid()
            vals_each_dim[design_var.name] = vals

        # meshgrid each dimension's results and combine them all
        vals_grid = np.meshgrid(*vals_each_dim.values())
        vals_grid = (np.ravel(x).tolist() for x in vals_grid)
        return dict(zip(vals_each_dim.keys(), vals_grid))


class AbstractMethodRandom(MethodIndependent, ABC):
    """Select parameters with an object with a ``random`` method."""

    num_points: pd.PositiveInt = pd.Field(
        ...,
        title="Number of Points",
        description="Maximum number of sampling points to perform in the sweep.",
    )

    @abstractmethod
    def get_sampler(self, parameters: Tuple[ParameterType, ...]) -> qmc.QMCEngine:
        """Sampler for this ``Method`` class. If ``None``, sets a default."""

    def sample(self, parameters: Tuple[ParameterType, ...], **kwargs) -> Dict[str, Any]:
        """Defines how the design parameters are sampled on grid."""

        sampler = self.get_sampler(parameters)
        pts_01 = sampler.random(self.num_points)

        # for each dimension, sample `num_points` points and combine them all
        result = {}
        for i, design_var in enumerate(parameters):
            pts_i_01 = pts_01[..., i]
            values = design_var.select_from_01(pts_i_01)
            result[design_var.name] = values
        return result


class MethodMonteCarlo(AbstractMethodRandom):
    """Select sampling points using Monte Carlo sampling (Latin Hypercube method).

    Example
    -------
    >>> import tidy3d.plugins.design as tdd
    >>> method = tdd.MethodMonteCarlo(num_points=20)
    """

    def get_sampler(self, parameters: Tuple[ParameterType, ...]) -> qmc.QMCEngine:
        """Sampler for this ``Method`` class."""

        d = len(parameters)
        return DEFAULT_MONTE_CARLO_SAMPLER_TYPE(d=d)


class MethodRandom(AbstractMethodRandom):
    """Select sampling points uniformly at random.

    Example
    -------
    >>> import tidy3d.plugins.design as tdd
    >>> method = tdd.MethodRandom(num_points=20, monte_carlo_warning=False)
    """

    monte_carlo_warning: bool = pd.Field(
        True,
        title="Monte Carlo Suggestion",
        description="We recommend you use ``MethodMonteCarlo`` as it is more efficient at sampling."
        " Setting this field to ``False`` will disable the "
        "warning that occurs when this class is made.",
    )

    @pd.validator("monte_carlo_warning", always=True)
    def _suggest_monte_carlo(cls, val):
        """Suggest that the user use ``MethodMonteCarlo`` instead of this method."""
        if val:
            log.warning(
                "We recommend using the Monte Carlo method to sample your design space instead of "
                "this method, which samples uniformly at random. Monte Carlo is more efficient at "
                "sampling and generally needs fewer points than uniform random sampling. "
                "Please consider using 'sweep.MethodMonteCarlo'. "
                "If you are intentionally using uniform random sampling, "
                "you can disable this warning by setting 'monte_carlo_warning=False' "
                "in 'MethodRandom'."
            )
        return val

    def get_sampler(self, parameters: Tuple[ParameterType, ...]) -> qmc.QMCEngine:
        """Sampler for this ``Method`` class."""

        d = len(parameters)

        class UniformRandomSampler:
            """Has ``.random(n)`` returning ``(n, d)`` array sampled random uniformly in [0, 1]."""

            def random(self, n) -> np.ndarray:
                """Return ``(n, d)``-shaped array sampled uniformly at random in range [0, 1]."""
                return np.random.random((n, d))

        return UniformRandomSampler()


class MethodRandomCustom(AbstractMethodRandom):
    """Select parameters with an object with a user supplied sampler with a ``.random`` method.

    Example
    -------
    >>> import tidy3d.plugins.design as tdd
    >>> import scipy.stats.qmc as qmc
    >>> sampler = qmc.Halton(d=3)
    >>> method = tdd.MethodRandomCustom(num_points=20, sampler=sampler)
    """

    sampler: Any = pd.Field(
        None,
        title="Custom Sampler",
        description="An object with a ``.random(n)`` method, which returns a ``np.ndarray`` "
        "of shape ``(n, d)`` where d is the number of dimensions of the design space. "
        "Values must lie between [0, 1] and will be re-scaled depending on the design parameters. "
        " Compatible objects include instances of ``scipy.stats.qmc.QMCEngine``, but other objects "
        " can also be supplied.",
    )

    @pd.validator("sampler")
    def _check_sampler(cls, val):
        """make sure sampler has required methods."""
        if not hasattr(val, "random"):
            raise ValueError(
                "Sampler must have a 'random(n)' method, "
                "returning a numpy array of shape '(n, d)' where 'd' is the number of dimensions "
                "in the design space."
            )
        n = 30
        sample_values = val.random(n)
        if not isinstance(sample_values, np.ndarray):
            raise ValueError(
                f"'sampler.random(n)' must return a 'np.ndarray' object, got {type(sample_values)}."
            )
        sample_shape = sample_values.shape
        if len(sample_shape) != 2:
            raise ValueError(
                f"The 'sampler.random(n)' method must give a 'np.ndarray of shape (n, d)', "
                "where 'd' is the number of dimensions in the design parameters. "
                f"Supplied sampler gave an array with {len(sample_shape)} dimensions."
            )
        if sample_shape[0] != n:
            raise ValueError(
                f"The 'sampler.random(n)' method must give a 'np.ndarray of shape (n, d)', "
                "where 'd' is the number of dimensions in the design parameters. "
                f"Supplied sampler gave an array of shape ({sample_shape[0]}, d)."
            )
        if np.any(sample_values > 1) or np.any(sample_values < 0):
            raise ValueError(
                "The 'sampler.random(n)' method must give a 'np.ndarray of shape (n, d)' where all "
                "values lie between 0 and 1. After the points are generated, "
                "their values will be resampled appropriately depending "
                "on the 'parameters' used in the parameter sweep 'DesignSpace'."
            )
        return val

    def get_sampler(self, parameters: Tuple[ParameterType, ...]) -> qmc.QMCEngine:
        """Sampler for this ``Method`` class. If ``None``, sets a default."""

        num_dims_vars = len(parameters)
        num_dims_sampler = self.sampler.random(1).size

        if num_dims_sampler != num_dims_vars:
            raise ValueError(
                f"The sampler {self.sampler} has {num_dims_sampler} dimensions, "
                f"but the design space has {num_dims_vars} dimensions. These must be equivalent. "
            )

        return self.sampler


MethodType = Union[MethodMonteCarlo, MethodGrid, MethodRandom, MethodRandomCustom]



"""Fit PoleResidue Dispersion models to optical NK data based on web service"""

from __future__ import annotations

import ssl
from enum import Enum
from typing import Optional, Tuple

import pydantic.v1 as pydantic
import requests
from pydantic.v1 import Field, NonNegativeFloat, PositiveFloat, PositiveInt, validator

from tidy3d.web.core.environment import Env
from tidy3d.web.core.http_util import get_headers

from ...components.base import Tidy3dBaseModel, skip_if_fields_missing
from ...components.medium import PoleResidue
from ...components.types import Literal
from ...constants import HERTZ, MICROMETER
from ...exceptions import SetupError, Tidy3dError, WebError
from ...log import log
from .fit import DispersionFitter

BOUND_MAX_FACTOR = 10

URL_ENV = {
    "local": "http://127.0.0.1:8000",
    "dev": "https://tidy3d-service.dev-simulation.cloud",
    "prod": "https://tidy3d-service.simulation.cloud",
}


class ExceptionCodes(Enum):
    """HTTP exception codes to handle individually."""

    GATEWAY_TIMEOUT = 504
    NOT_FOUND = 404


class AdvancedFitterParam(Tidy3dBaseModel):
    """Advanced fitter parameters"""

    bound_amp: NonNegativeFloat = Field(
        None,
        title="Upper bound of oscillator strength",
        description="Upper bound of real and imagniary part of oscillator "
        "strength ``c`` in the model :class:`.PoleResidue` (The default 'None' will trigger "
        "automatic setup based on the frequency range of interest).",
        units=HERTZ,
    )
    bound_f: NonNegativeFloat = Field(
        None,
        title="Upper bound of pole frequency",
        description="Upper bound of real and imaginary part of ``a`` that corresponds to pole "
        "damping rate and frequency in the model :class:`.PoleResidue` (The default 'None' "
        "will trigger automatic setup based on the frequency range of interest).",
        units=HERTZ,
    )
    bound_f_lower: NonNegativeFloat = Field(
        0.0,
        title="Lower bound of pole frequency",
        description="Lower bound of imaginary part of ``a`` that corresponds to pole "
        "frequency in the model :class:`.PoleResidue`.",
        units=HERTZ,
    )
    bound_eps_inf: float = Field(
        10.0,
        title="Upper bound of epsilon at infinity frequency",
        description="Upper bound of epsilon at infinity frequency. It must be no less than 1.",
        ge=1,
    )
    constraint: Literal["hard", "soft"] = Field(
        "hard",
        title="Type of constraint for stability",
        description="Stability constraint: 'hard' constraints are generally recommended since "
        "they are faster to compute per iteration, and they often require fewer iterations to "
        "converge since the search space is smaller. But sometimes the search space is "
        "so restrictive that all good solutions are missed, then please try the 'soft' constraints "
        "for larger search space. However, both constraints improve stability equally well.",
    )
    nlopt_maxeval: PositiveInt = Field(
        5000,
        title="Number of inner iterations",
        description="Number of iterations in each inner optimization.",
    )
    random_seed: Optional[int] = Field(
        0,
        title="Random seed for starting coefficients",
        description="The fitting tool performs global optimizations with random "
        "starting coefficients. With the same random seed, one obtains identical "
        "results when re-running the fitter; on the other hand, if "
        "one wants to re-run the fitter several times to obtain the best results, "
        "the value of the seed should be changed, or set to  ``None`` so that "
        "the starting coefficients are different each time. ",
        ge=0,
        lt=2**32,
    )

    @validator("bound_f_lower", always=True)
    @skip_if_fields_missing(["bound_f"])
    def _validate_lower_frequency_bound(cls, val, values):
        """bound_f_lower cannot be larger than bound_f."""
        if values["bound_f"] is not None and val > values["bound_f"]:
            raise SetupError(
                "The upper bound 'bound_f' cannot be smaller "
                "than the lower bound 'bound_f_lower'."
            )
        return val


class FitterData(AdvancedFitterParam):
    """Data class for request body of Fitter where dipsersion data is input through tuple."""

    wvl_um: Tuple[float, ...] = Field(
        ...,
        title="Wavelengths",
        description="A set of wavelengths for dispersion data.",
        units=MICROMETER,
    )
    n_data: Tuple[float, ...] = Field(
        ...,
        title="Index of refraction",
        description="Real part of the complex index of refraction at each wavelength.",
    )
    k_data: Tuple[float, ...] = Field(
        None,
        title="Extinction coefficient",
        description="Imaginary part of the complex index of refraction at each wavelength.",
    )
    num_poles: PositiveInt = Field(
        1, title="Number of poles", description="Number of poles in model."
    )
    num_tries: PositiveInt = Field(
        50,
        title="Number of tries",
        description="Number of optimizations to run with different initial guess.",
    )
    tolerance_rms: NonNegativeFloat = Field(
        0.0,
        title="RMS error tolerance",
        description="RMS error below which the fit is successful and result is returned.",
    )
    bound_amp: PositiveFloat = Field(
        100.0,
        title="Upper bound of oscillator strength",
        description="Upper bound of oscillator strength in the model.",
        units="eV",
    )
    bound_f: PositiveFloat = Field(
        100.0,
        title="Upper bound of pole frequency",
        description="Upper bound of pole frequency in the model.",
        units="eV",
    )

    @staticmethod
    def create(
        fitter: DispersionFitter,
        num_poles: PositiveInt,
        num_tries: PositiveInt,
        tolerance_rms: NonNegativeFloat,
        advanced_param: AdvancedFitterParam,
    ) -> FitterData:
        """Setup FitterData to be provided to web service

        Parameters
        ----------
        fitter : DispersionFitter
            Fitter with the data to fit.
        num_poles : PositiveInt
            Number of poles in the model.
        num_tries : PositiveInt
            Number of optimizations to run with random initial guess.
        tolerance_rms : NonNegativeFloat
            RMS error below which the fit is successful and the result is returned.
        advanced_param : :class:`AdvancedFitterParam`
            Other advanced parameters.

        Returns
        -------
        :class:`FitterData`
            Data class for request body of Fitter where dispersion
            data is input through tuple.
        """

        # set up bound_f, bound_amp
        if advanced_param.bound_f is None:
            new_bound_f = (
                advanced_param.bound_f_lower + fitter.frequency_range[1] * BOUND_MAX_FACTOR
            )
            advanced_param = advanced_param.copy(update={"bound_f": new_bound_f})
        if advanced_param.bound_amp is None:
            new_bound_amp = fitter.frequency_range[1] * BOUND_MAX_FACTOR
            advanced_param = advanced_param.copy(update={"bound_amp": new_bound_amp})

        wvl_um, n_data, k_data = fitter.data_in_range

        if fitter.lossy:
            k_data = k_data.tolist()
        else:
            k_data = None

        task = FitterData(
            wvl_um=wvl_um.tolist(),
            n_data=n_data.tolist(),
            k_data=k_data,
            num_poles=num_poles,
            num_tries=num_tries,
            tolerance_rms=tolerance_rms,
            bound_amp=fitter._Hz_to_eV(advanced_param.bound_amp),
            bound_f=fitter._Hz_to_eV(advanced_param.bound_f),
            bound_f_lower=fitter._Hz_to_eV(advanced_param.bound_f_lower),
            bound_eps_inf=advanced_param.bound_eps_inf,
            constraint=advanced_param.constraint,
            nlopt_maxeval=advanced_param.nlopt_maxeval,
            random_seed=advanced_param.random_seed,
        )
        return task

    @staticmethod
    def _set_url(config_env: Literal["default", "dev", "prod", "local"] = "default"):
        """Set the url of python web service

        Parameters
        ----------
        config_env : Literal["default", "dev", "prod", "local"], optional
            Service environment to pick from
        """

        _env = config_env
        if _env == "default":
            _env = "dev" if "dev" in Env.current.web_api_endpoint else "prod"
        return URL_ENV[_env]

    @staticmethod
    def _setup_server(url_server: str):
        """set up web server access

        Parameters
        ----------
        url_server : str
            URL for the server
        """

        try:
            # test connection
            resp = requests.get(f"{url_server}/health", verify=Env.current.ssl_verify)
            resp.raise_for_status()
        except (requests.exceptions.SSLError, ssl.SSLError):
            log.info("Retrying with SSL verification disabled.")
            Env.current.ssl_verify = False
            resp = requests.get(f"{url_server}/health", verify=Env.current.ssl_verify)
        except Exception as e:
            raise WebError("Connection to the server failed. Please try again.") from e

        return get_headers()

    def run(self) -> Tuple[PoleResidue, float]:
        """Execute the data fit using the stable fitter in the server.

        Returns
        -------
        Tuple[:class:`.PoleResidue`, float]
            Best results of multiple fits: (dispersive medium, RMS error).
        """

        url_server = self._set_url("default")
        headers = self._setup_server(url_server)

        resp = requests.post(
            f"{url_server}/dispersion/fit",
            headers=headers,
            data=self.json(),
            verify=Env.current.ssl_verify,
        )

        try:
            resp.raise_for_status()
        except Exception as e:
            if resp.status_code == ExceptionCodes.GATEWAY_TIMEOUT.value:
                msg = (
                    (
                        "Fitter failed due to timeout. Try to decrease the number of tries or "
                        "inner iterations, to relax the RMS tolerance, or to use the 'hard' "
                        "constraint."
                    )
                    if self.constraint != "hard"
                    else (
                        "Fitter failed due to timeout. Try to decrease the number of tries or "
                        "inner iterations, or to relax the RMS tolerance."
                    )
                )
                raise Tidy3dError(msg) from e

            raise WebError(
                "Fitter failed. Try again, tune the parameters, or contact us for more help."
            ) from e

        run_result = resp.json()
        best_medium = PoleResidue.parse_raw(run_result["message"])
        best_rms = float(run_result["rms"])

        if best_rms < self.tolerance_rms:
            log.info("Found optimal fit with RMS error %.3g", best_rms)
        else:
            log.warning(
                "Unable to fit with RMS error under 'tolerance_rms' of %.3g", self.tolerance_rms
            )
            log.info("Returning best fit with RMS error %.3g", best_rms)

        return best_medium, best_rms


def run(
    fitter: DispersionFitter,
    num_poles: PositiveInt = 1,
    num_tries: PositiveInt = 50,
    tolerance_rms: NonNegativeFloat = 1e-2,
    advanced_param: AdvancedFitterParam = AdvancedFitterParam(),
) -> Tuple[PoleResidue, float]:
    """Execute the data fit using the stable fitter in the server.

    Parameters
    ----------
    fitter : DispersionFitter
        Fitter with the data to fit.
    num_poles : PositiveInt, optional
        Number of poles in the model.
    num_tries : PositiveInt, optional
        Number of optimizations to run with random initial guess.
    tolerance_rms : NonNegativeFloat, optional
        RMS error below which the fit is successful and the result is returned.
    advanced_param : :class:`AdvancedFitterParam`, optional
        Advanced parameters passed on to the server.

    Returns
    -------
    Tuple[:class:`.PoleResidue`, float]
        Best results of multiple fits: (dispersive medium, RMS error).
    """
    task = FitterData.create(fitter, num_poles, num_tries, tolerance_rms, advanced_param)
    return task.run()


class StableDispersionFitter(DispersionFitter):
    """Deprecated."""

    @pydantic.root_validator()
    def _deprecate_stable_fitter(cls, values):
        log.warning(
            "'StableDispersionFitter' has been deprecated. Use 'DispersionFitter' with "
            "'tidy3d.plugins.dispersion.web.run' to access the stable fitter from the web server."
        )
        return values

    def fit(
        self,
        num_poles: PositiveInt = 1,
        num_tries: PositiveInt = 50,
        tolerance_rms: NonNegativeFloat = 1e-2,
        guess: PoleResidue = None,
        advanced_param: AdvancedFitterParam = AdvancedFitterParam(),
    ) -> Tuple[PoleResidue, float]:
        """Deprecated."""
        return run(self, num_poles, num_tries, tolerance_rms, advanced_param)



"""Tidy3d core log, need init config from Tidy3d api"""

import logging as log

# default setting
config_setting = {
    "logger": log,
    "logger_console": None,
    "version": "",
}


def set_config(logger, logger_console, version: str):
    """Init tidy3d core logger and logger console.

    Parameters
    ----------
    logger : :class:`.Logger`
        Tidy3d log Logger.
    logger_console : :class:`.Console`
        Get console from logging handlers.
    version : str
        tidy3d version
    """
    config_setting["logger"] = logger
    config_setting["logger_console"] = logger_console
    config_setting["version"] = version


def get_logger():
    """Get logging handlers."""
    return config_setting["logger"]


def get_logger_console():
    """Get console from logging handlers."""
    return config_setting["logger_console"]


def get_version():
    """Get version from cache."""
    return config_setting["version"]



"""Abstract bases for classes that define how data is recorded from simulation."""

from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np
import pydantic.v1 as pd

from ..base import cached_property
from ..geometry.base import Box
from ..types import ArrayFloat1D, Axis, Numpy
from ..viz import PlotParams, plot_params_monitor


class AbstractMonitor(Box, ABC):
    """Abstract base class for steady-state monitors."""

    name: str = pd.Field(
        ...,
        title="Name",
        description="Unique name for monitor.",
        min_length=1,
    )

    @cached_property
    def plot_params(self) -> PlotParams:
        """Default parameters for plotting a Monitor object."""
        return plot_params_monitor

    @cached_property
    def geometry(self) -> Box:
        """:class:`Box` representation of monitor.

        Returns
        -------
        :class:`Box`
            Representation of the monitor geometry as a :class:`Box`.
        """
        return Box(center=self.center, size=self.size)

    @abstractmethod
    def storage_size(self, num_cells: int, tmesh: ArrayFloat1D) -> int:
        """Size of monitor storage given the number of points after discretization.

        Parameters
        ----------
        num_cells : int
            Number of grid cells within the monitor after discretization by a :class:`Simulation`.
        tmesh : Array
            The discretized time mesh of a :class:`Simulation`.

        Returns
        -------
        int
            Number of bytes to be stored in monitor.
        """

    def downsample(self, arr: Numpy, axis: Axis) -> Numpy:
        """Downsample a 1D array making sure to keep the first and last entries, based on the
        spatial interval defined for the ``axis``.

        Parameters
        ----------
        arr : Numpy
            A 1D array of arbitrary type.
        axis : Axis
            Axis for which to select the interval_space defined for the monitor.

        Returns
        -------
        Numpy
            Downsampled array.
        """

        size = len(arr)
        interval = self.interval_space[axis]
        # There should always be at least 3 indices for "surface" monitors. Also, if the
        # size along this dim is already smaller than the interval, then don't downsample.
        if size < 4 or (size - 1) <= interval:
            return arr
        # make sure the last index is always included
        inds = np.arange(0, size, interval)
        if inds[-1] != size - 1:
            inds = np.append(inds, size - 1)
        return arr[inds]

    def downsampled_num_cells(self, num_cells: Tuple[int, int, int]) -> Tuple[int, int, int]:
        """Given a tuple of the number of cells spanned by the monitor along each dimension,
        return the number of cells one would have after downsampling based on ``interval_space``.
        """
        arrs = [np.arange(ncells) for ncells in num_cells]
        return tuple((self.downsample(arr, axis=dim).size for dim, arr in enumerate(arrs)))
